/*
 * livepatch_bsc1227472
 *
 * Fix for CVE-2021-47600, bsc#1227472
 *
 *  Upstream commit:
 *  1b8d2789dad0 ("dm btree remove: fix use after free in rebalance_children()")
 *
 *  SLE12-SP5 commit:
 *  bba5423b0e8029e02e318f52d88fa6a117e60ce1
 *
 *  SLE15-SP2 and -SP3 commit:
 *  088e9db2f3891442c6d1d6c51660f01051ed31dc
 *
 *  SLE15-SP4 and -SP5 commit:
 *  abb14bce9cbeccfab3cd1ea831f01885dab7e01b
 *
 *  SLE15-SP6 commit:
 *  Not affected
 *
 *  Copyright (c) 2024 SUSE
 *  Author: Marcos Paulo de Souza <mpdesouza@suse.com>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/* klp-ccp: from drivers/md/persistent-data/dm-block-manager.h */
#include <linux/types.h>
#include <linux/blkdev.h>

typedef uint64_t dm_block_t;
struct dm_block;

static dm_block_t (*klpe_dm_block_location)(struct dm_block *b);
static void *(*klpe_dm_block_data)(struct dm_block *b);

struct dm_block_manager;

static unsigned (*klpe_dm_bm_block_size)(struct dm_block_manager *bm);

struct dm_block_validator {
	const char *name;
	void (*prepare_for_write)(struct dm_block_validator *v, struct dm_block *b, size_t block_size);

	/*
	 * Return 0 if the checksum is valid or < 0 on error.
	 */
	int (*check)(struct dm_block_validator *v, struct dm_block *b, size_t block_size);
};

/* klp-ccp: from drivers/md/persistent-data/dm-btree.h */
struct dm_btree_value_type {
	void *context;

	/*
	 * The size in bytes of each value.
	 */
	uint32_t size;

	/*
	 * Any of these methods can be safely set to NULL if you do not
	 * need the corresponding feature.
	 */

	/*
	 * The btree is making a duplicate of the value, for instance
	 * because previously-shared btree nodes have now diverged.
	 * @value argument is the new copy that the copy function may modify.
	 * (Probably it just wants to increment a reference count
	 * somewhere.) This method is _not_ called for insertion of a new
	 * value: It is assumed the ref count is already 1.
	 */
	void (*inc)(void *context, const void *value);

	/*
	 * This value is being deleted.  The btree takes care of freeing
	 * the memory pointed to by @value.  Often the del function just
	 * needs to decrement a reference count somewhere.
	 */
	void (*dec)(void *context, const void *value);

	/*
	 * A test for equality between two values.  When a value is
	 * overwritten with a new one, the old one has the dec method
	 * called _unless_ the new and old value are deemed equal.
	 */
	int (*equal)(void *context, const void *value1, const void *value2);
};

struct dm_btree_info {
	struct dm_transaction_manager *tm;

	/*
	 * Number of nested btrees. (Not the depth of a single tree.)
	 */
	unsigned levels;
	struct dm_btree_value_type value_type;
};

/* klp-ccp: from drivers/md/persistent-data/dm-btree-internal.h */
struct node_header {
	__le32 csum;
	__le32 flags;
	__le64 blocknr; /* Block this node is supposed to live in. */

	__le32 nr_entries;
	__le32 max_entries;
	__le32 value_size;
	__le32 padding;
} __packed;

struct btree_node {
	struct node_header header;
	__le64 keys[0];
} __packed;

static void (*klpe_inc_children)(struct dm_transaction_manager *tm, struct btree_node *n,
		  struct dm_btree_value_type *vt);

struct shadow_spine;

static struct dm_block *(*klpe_shadow_current)(struct shadow_spine *s);

static inline __le64 *key_ptr(struct btree_node *n, uint32_t index)
{
	return n->keys + index;
}

static inline void *value_base(struct btree_node *n)
{
	return &n->keys[le32_to_cpu(n->header.max_entries)];
}

static inline void *value_ptr(struct btree_node *n, uint32_t index)
{
	uint32_t value_size = le32_to_cpu(n->header.value_size);
	return value_base(n) + (value_size * index);
}

static inline uint64_t value64(struct btree_node *n, uint32_t index)
{
	__le64 *values_le = value_base(n);

	return le64_to_cpu(values_le[index]);
}

static int (*klpe_lower_bound)(struct btree_node *n, uint64_t key);

static struct dm_block_validator (*klpe_btree_node_validator);

/* klp-ccp: from drivers/md/persistent-data/dm-transaction-manager.h */
static int (*klpe_dm_tm_shadow_block)(struct dm_transaction_manager *tm, dm_block_t orig,
		       struct dm_block_validator *v,
		       struct dm_block **result, int *inc_children);

static int (*klpe_dm_tm_read_lock)(struct dm_transaction_manager *tm, dm_block_t b,
		    struct dm_block_validator *v,
		    struct dm_block **result);

static void (*klpe_dm_tm_unlock)(struct dm_transaction_manager *tm, struct dm_block *b);

static void (*klpe_dm_tm_dec)(struct dm_transaction_manager *tm, dm_block_t b);

static struct dm_block_manager *(*klpe_dm_tm_get_bm)(struct dm_transaction_manager *tm);

/* klp-ccp: from drivers/md/persistent-data/dm-btree-remove.c */
#include <linux/export.h>

static void (*klpe_node_shift)(struct btree_node *n, int shift);

static void (*klpe_node_copy)(struct btree_node *left, struct btree_node *right, int shift);

static void (*klpe_delete_at)(struct btree_node *n, unsigned index);

static unsigned merge_threshold(struct btree_node *n)
{
	return le32_to_cpu(n->header.max_entries) / 3;
}

struct child {
	unsigned index;
	struct dm_block *block;
	struct btree_node *n;
};

static int klpr_init_child(struct dm_btree_info *info, struct dm_btree_value_type *vt,
		      struct btree_node *parent,
		      unsigned index, struct child *result)
{
	int r, inc;
	dm_block_t root;

	result->index = index;
	root = value64(parent, index);

	r = (*klpe_dm_tm_shadow_block)(info->tm, root, &(*klpe_btree_node_validator),
			       &result->block, &inc);
	if (r)
		return r;

	result->n = (*klpe_dm_block_data)(result->block);

	if (inc)
		(*klpe_inc_children)(info->tm, result->n, vt);

	*((__le64 *) value_ptr(parent, index)) =
		cpu_to_le64((*klpe_dm_block_location)(result->block));

	return 0;
}

static void klpr_exit_child(struct dm_btree_info *info, struct child *c)
{
	(*klpe_dm_tm_unlock)(info->tm, c->block);
}

static void (*klpe_shift)(struct btree_node *left, struct btree_node *right, int count);

static void klpr___rebalance2(struct dm_btree_info *info, struct btree_node *parent,
			 struct child *l, struct child *r)
{
	struct btree_node *left = l->n;
	struct btree_node *right = r->n;
	uint32_t nr_left = le32_to_cpu(left->header.nr_entries);
	uint32_t nr_right = le32_to_cpu(right->header.nr_entries);
	/*
	 * Ensure the number of entries in each child will be greater
	 * than or equal to (max_entries / 3 + 1), so no matter which
	 * child is used for removal, the number will still be not
	 * less than (max_entries / 3).
	 */
	unsigned int threshold = 2 * (merge_threshold(left) + 1);

	if (nr_left + nr_right < threshold) {
		/*
		 * Merge
		 */
		(*klpe_node_copy)(left, right, -nr_right);
		left->header.nr_entries = cpu_to_le32(nr_left + nr_right);
		(*klpe_delete_at)(parent, r->index);

		/*
		 * We need to decrement the right block, but not it's
		 * children, since they're still referenced by left.
		 */
		(*klpe_dm_tm_dec)(info->tm, (*klpe_dm_block_location)(r->block));
	} else {
		/*
		 * Rebalance.
		 */
		unsigned target_left = (nr_left + nr_right) / 2;
		(*klpe_shift)(left, right, nr_left - target_left);
		*key_ptr(parent, r->index) = right->keys[0];
	}
}

static int (*klpe_rebalance2)(struct shadow_spine *s, struct dm_btree_info *info,
		      struct dm_btree_value_type *vt, unsigned left_index);

static void klpr_delete_center_node(struct dm_btree_info *info, struct btree_node *parent,
			       struct child *l, struct child *c, struct child *r,
			       struct btree_node *left, struct btree_node *center, struct btree_node *right,
			       uint32_t nr_left, uint32_t nr_center, uint32_t nr_right)
{
	uint32_t max_entries = le32_to_cpu(left->header.max_entries);
	unsigned shift = min(max_entries - nr_left, nr_center);

	BUG_ON(nr_left + shift > max_entries);
	(*klpe_node_copy)(left, center, -shift);
	left->header.nr_entries = cpu_to_le32(nr_left + shift);

	if (shift != nr_center) {
		shift = nr_center - shift;
		BUG_ON((nr_right + shift) > max_entries);
		(*klpe_node_shift)(right, shift);
		(*klpe_node_copy)(center, right, shift);
		right->header.nr_entries = cpu_to_le32(nr_right + shift);
	}
	*key_ptr(parent, r->index) = right->keys[0];

	(*klpe_delete_at)(parent, c->index);
	r->index--;

	(*klpe_dm_tm_dec)(info->tm, (*klpe_dm_block_location)(c->block));
	klpr___rebalance2(info, parent, l, r);
}

static void klpr_redistribute3(struct dm_btree_info *info, struct btree_node *parent,
			  struct child *l, struct child *c, struct child *r,
			  struct btree_node *left, struct btree_node *center, struct btree_node *right,
			  uint32_t nr_left, uint32_t nr_center, uint32_t nr_right)
{
	int s;
	uint32_t max_entries = le32_to_cpu(left->header.max_entries);
	unsigned total = nr_left + nr_center + nr_right;
	unsigned target_right = total / 3;
	unsigned remainder = (target_right * 3) != total;
	unsigned target_left = target_right + remainder;

	BUG_ON(target_left > max_entries);
	BUG_ON(target_right > max_entries);

	if (nr_left < nr_right) {
		s = nr_left - target_left;

		if (s < 0 && nr_center < -s) {
			/* not enough in central node */
			(*klpe_shift)(left, center, -nr_center);
			s += nr_center;
			(*klpe_shift)(left, right, s);
			nr_right += s;
		} else
			(*klpe_shift)(left, center, s);

		(*klpe_shift)(center, right, target_right - nr_right);

	} else {
		s = target_right - nr_right;
		if (s > 0 && nr_center < s) {
			/* not enough in central node */
			(*klpe_shift)(center, right, nr_center);
			s -= nr_center;
			(*klpe_shift)(left, right, s);
			nr_left -= s;
		} else
			(*klpe_shift)(center, right, s);

		(*klpe_shift)(left, center, nr_left - target_left);
	}

	*key_ptr(parent, c->index) = center->keys[0];
	*key_ptr(parent, r->index) = right->keys[0];
}

static void klpr___rebalance3(struct dm_btree_info *info, struct btree_node *parent,
			 struct child *l, struct child *c, struct child *r)
{
	struct btree_node *left = l->n;
	struct btree_node *center = c->n;
	struct btree_node *right = r->n;

	uint32_t nr_left = le32_to_cpu(left->header.nr_entries);
	uint32_t nr_center = le32_to_cpu(center->header.nr_entries);
	uint32_t nr_right = le32_to_cpu(right->header.nr_entries);

	unsigned threshold = merge_threshold(left) * 4 + 1;

	BUG_ON(left->header.max_entries != center->header.max_entries);
	BUG_ON(center->header.max_entries != right->header.max_entries);

	if ((nr_left + nr_center + nr_right) < threshold)
		klpr_delete_center_node(info, parent, l, c, r, left, center, right,
				   nr_left, nr_center, nr_right);
	else
		klpr_redistribute3(info, parent, l, c, r, left, center, right,
			      nr_left, nr_center, nr_right);
}

static int klpr_rebalance3(struct shadow_spine *s, struct dm_btree_info *info,
		      struct dm_btree_value_type *vt, unsigned left_index)
{
	int r;
	struct btree_node *parent = (*klpe_dm_block_data)((*klpe_shadow_current)(s));
	struct child left, center, right;

	/*
	 * FIXME: fill out an array?
	 */
	r = klpr_init_child(info, vt, parent, left_index, &left);
	if (r)
		return r;

	r = klpr_init_child(info, vt, parent, left_index + 1, &center);
	if (r) {
		klpr_exit_child(info, &left);
		return r;
	}

	r = klpr_init_child(info, vt, parent, left_index + 2, &right);
	if (r) {
		klpr_exit_child(info, &left);
		klpr_exit_child(info, &center);
		return r;
	}

	klpr___rebalance3(info, parent, &left, &center, &right);

	klpr_exit_child(info, &left);
	klpr_exit_child(info, &center);
	klpr_exit_child(info, &right);

	return 0;
}

int klpp_rebalance_children(struct shadow_spine *s,
			      struct dm_btree_info *info,
			      struct dm_btree_value_type *vt, uint64_t key)
{
	int i, r, has_left_sibling, has_right_sibling;
	struct btree_node *n;

	n = (*klpe_dm_block_data)((*klpe_shadow_current)(s));

	if (le32_to_cpu(n->header.nr_entries) == 1) {
		struct dm_block *child;
		dm_block_t b = value64(n, 0);

		r = (*klpe_dm_tm_read_lock)(info->tm, b, &(*klpe_btree_node_validator), &child);
		if (r)
			return r;

		memcpy(n, (*klpe_dm_block_data)(child),
		       (*klpe_dm_bm_block_size)((*klpe_dm_tm_get_bm)(info->tm)));

		(*klpe_dm_tm_dec)(info->tm, (*klpe_dm_block_location)(child));
		(*klpe_dm_tm_unlock)(info->tm, child);
		return 0;
	}

	i = (*klpe_lower_bound)(n, key);
	if (i < 0)
		return -ENODATA;

	has_left_sibling = i > 0;
	has_right_sibling = i < (le32_to_cpu(n->header.nr_entries) - 1);

	if (!has_left_sibling)
		r = (*klpe_rebalance2)(s, info, vt, i);

	else if (!has_right_sibling)
		r = (*klpe_rebalance2)(s, info, vt, i - 1);

	else
		r = klpr_rebalance3(s, info, vt, i - 1);

	return r;
}


#include "livepatch_bsc1227472.h"

#include <linux/kernel.h>
#include <linux/module.h>
#include "../kallsyms_relocs.h"

#define LP_MODULE "dm_persistent_data"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "btree_node_validator", (void *)&klpe_btree_node_validator,
	  "dm_persistent_data" },
	{ "delete_at", (void *)&klpe_delete_at, "dm_persistent_data" },
	{ "dm_block_data", (void *)&klpe_dm_block_data, "dm_persistent_data" },
	{ "dm_block_location", (void *)&klpe_dm_block_location,
	  "dm_persistent_data" },
	{ "dm_bm_block_size", (void *)&klpe_dm_bm_block_size,
	  "dm_persistent_data" },
	{ "dm_tm_dec", (void *)&klpe_dm_tm_dec, "dm_persistent_data" },
	{ "dm_tm_get_bm", (void *)&klpe_dm_tm_get_bm, "dm_persistent_data" },
	{ "dm_tm_read_lock", (void *)&klpe_dm_tm_read_lock,
	  "dm_persistent_data" },
	{ "dm_tm_shadow_block", (void *)&klpe_dm_tm_shadow_block,
	  "dm_persistent_data" },
	{ "dm_tm_unlock", (void *)&klpe_dm_tm_unlock, "dm_persistent_data" },
	{ "inc_children", (void *)&klpe_inc_children, "dm_persistent_data" },
	{ "lower_bound", (void *)&klpe_lower_bound, "dm_persistent_data" },
	{ "node_copy", (void *)&klpe_node_copy, "dm_persistent_data" },
	{ "node_shift", (void *)&klpe_node_shift, "dm_persistent_data" },
	{ "rebalance2", (void *)&klpe_rebalance2, "dm_persistent_data" },
	{ "shadow_current", (void *)&klpe_shadow_current,
	  "dm_persistent_data" },
	{ "shift", (void *)&klpe_shift, "dm_persistent_data" },
};

static int module_notify(struct notifier_block *nb,
			unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, LP_MODULE))
		return 0;
	mutex_lock(&module_mutex);
	ret = __klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
	mutex_unlock(&module_mutex);

	WARN(ret, "%s: delayed kallsyms lookup failed. System is broken and can crash.\n",
		__func__);

	return ret;
}

static struct notifier_block module_nb = {
	.notifier_call = module_notify,
	.priority = INT_MIN+1,
};

int livepatch_bsc1227472_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(LP_MODULE)) {
		ret = __klp_resolve_kallsyms_relocs(klp_funcs,
						    ARRAY_SIZE(klp_funcs));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void livepatch_bsc1227472_cleanup(void)
{
	unregister_module_notifier(&module_nb);
}
